/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/* 
 * File:   Converter.cpp
 * Author: tg
 * 
 * Created on July 12, 2016, 4:48 PM
 */

#include "Converter.h"
#include "Image.h"
#include "io.h"
#include "myExceptions.h"
#include "defines.h"

#include <sstream>
#include <iostream>
#include <cmath>
#include <iomanip>
#include <algorithm>
#include <stdexcept>

Converter::Converter(const std::string& inputdir, const std::string& basename, 
        const std::string& inroot, unsigned short from, unsigned short to, 
        bool verbose):
inputdir_ (inputdir), basename_(basename), inroot_ (inroot), from_(from), to_(to), 
        verbose_ (verbose){
    // make sure that inputdir ends with slash
    if (inputdir_.back() != '/' && inputdir_.back() != '\\') {
#ifdef _WIN32
        inputdir_.push_back ('\\');
#else   
        inputdir_.push_back('/');
#endif
        }
    }

/**
 * Convert infile to outfile. Currently a workaround while working on TIFF
 * files from PISA
 * @param sum
 */
Converter::Converter(const std::string& infile, const std::string& outfile, 
    bool verbose, const short crosscale) {
    std::vector<unsigned short> data;
    
    size_t peak;
    try {
    data = IO::in::tiff(infile.c_str(), width_, height_, verbose);
    }
    catch (const myExcepts::FileMissing& e) {
        data = IO::in::raw(infile.c_str(), width_, height_, peak, verbose);
    }
    Image<int> image(data, width_, height_, peak, crosscale);
    IO::out::cbf(outfile.c_str(), image);
}

/**
 * Carry out conversion
 * @param sum Sum this number of images before writing images. Useful for weekly 
 * diffracting samples to improve spot contrast
 */
void Converter::convert(unsigned short sum, const short crosscale) {
    std::vector<bool> conv_flags(to_ + 1, true);
    std::vector <unsigned short> f000x, f000y;
    f000x.reserve (256); // arbitrary size
    f000y.reserve (256); // arbitrary size

    for (size_t i = from_; i <= to_; i += sum) {

        std::string input = infile(i);
        try {
            size_t peak;
            // read first data image
            const std::vector<unsigned short> rawdata (IO::in::raw(input.c_str(), width_, height_, peak, verbose_));
            Image<int> image(rawdata, width_, height_, peak, crosscale);
            
            // calculate beam centre excluding pixel offsets
            const unsigned short peakx (peak % width_);
            f000x.push_back(peakx);
            const unsigned short peaky (peak / width_);
            f000y.push_back(peaky);


            // conversion including optional summation
            for (unsigned short s = 1; s < sum; ++s) {
                input = infile(i + s);
                const std::vector<unsigned short> rawdata(IO::in::raw(input.c_str(), width_, height_, peak, verbose_));
                Image <int> summand(rawdata, width_, height_, peak, crosscale);
                summand.add(image);

                const unsigned short peakx(peak % width_);
                f000x.push_back(peakx);
                const unsigned short peaky(peak / width_);
                f000y.push_back(peaky);
            }
            // set edges to -1
            // (image);
            size_t outidx = from_ + (i - from_) / sum;
            std::string output(outfile(basename_, outidx, to_));
            ++outidx;
            // write image data
            IO::out::cbf(output.c_str(), image);
            
        } catch (myExcepts::FileMissing& e) {
            conv_flags[i] = false;
        }
    }

    if (verbose_) {
        // use median of incident beam direction
        std::sort (f000x.begin(), f000x.end());
        std::sort (f000y.begin(), f000y.end());

        const size_t idx (f000x.size() / 2 + f000x.size()%2);

        float f000x_median;
        float f000y_median;
        
        px2mm(f000x[idx], f000y[idx], f000x_median, f000y_median);
        
        std::cout << "INCIDENT_BEAM_DIRECTION=" << f000x_median << " " << f000y_median << " FS ! " 
                << f000x[idx] << " " << f000y[idx] << "\n";
        std::cout << "DATA_RANGE= " << from_ << ' ' << from_ + (to_ - from_) / sum << '\n';
        print_missing(conv_flags);
    }
}

/**
 * Print segment information of XDS. The 512x512 quad is considered rigid and corrected
 * For the 1024x1024 Timepix installed at C-CINA, the relative orientation of these for
 * quads is decribed. According to Eric, top left and bottom right are parallel, with 
 * a small x- and larger y-translation. Similar for the top right and bottom right, but the 
 * pair is rotated by a small degree (-0.38 degree)
 * \param dx width of vertical gap between quads (mm)
 * \param dy width of horizontal gap between quads (mm)
 */
void Converter::segment() const {
    // ORGX and ORGY are set to width/2 and height/2, i.e. for the 1024x1024 chip,
    // the top left quad is not affected by the gap
    // print SEGMENT Information
    std::cout << "! XDS Segment information\n";
    // this should only be called for the large Timepix!
    for (unsigned short endx = 1; endx < width_; endx += 512) {
        float dorgx;
        // first pixel causes offest by 2x0.055
        short offset((endx - width_ / 2) / 256);
        switch (offset) {
            case -2: dorgx = -6.0;
                break;
            case -1: dorgx = -2.0;
                break;
            case 0: dorgx = 4.0;
                break;
            case 1: dorgx = 10.0;
                break;
            default:
                std::cout << "*** Error: Unknown Detector dimensions with endx = " 
                        << endx << " and offset = " << offset << ".\n";
                throw std::runtime_error("Offset Calculation: unknown dorgx");
        }
        if (endx > 512) {
            dorgx += Defs::Timepix::quad_gap_dx / Defs::Timepix::pixel_size_x;
        }

        for (unsigned short endy = 0; endy < width_; endy += 256) {
            float dorgy;
            // first pixel causes offest by 2x0.055
            short offset((endy - width_ / 2) / 256);
            switch (offset) {
                case -2: dorgy = -6.0;
                    break;
                case -1: dorgy = -2.0;
                    break;
                case 0: dorgy = 4.0;
                    break;
                case 1: dorgy = 10.0;
                    break;
                default:
                    std::cout << "*** Error: Unknown Detector dimensions.\n";
                    throw std::runtime_error("Offset Calculation: unknown dorgy");
            }
            if (endy > 512) {
                dorgy += Defs::Timepix::quad_gap_dy / Defs::Timepix::pixel_size_y;
            }

            // segments are 1-512, 1-512; 513-1024, 1-512;
            //              1-512, 513-1024; 513-1024
            std::cout
                    << "SEGMENT= " << endx + 2 << " " << endx + 254
                    << " " << endy + 2 << " " << endy + 254 << '\n'
                    << "DIRECTION_OF_SEGMENT_X-AXIS= 1.0 0.0 0.0\n"
                    << "DIRECTION_OF_SEGMENT_Y-AXIS= 0.0 1.0 0.0\n"
                    << "SEGMENT_ORGX=" << dorgx << '\n'
                    << "SEGMENT_ORGY=" << dorgy << '\n';
        }

    }
}
/** 
 * Convert pixels x/y into mm coordinates of the detector coordinate system
 * (useful for incident beam value)
 * @param x
 * @param y
 */
void Converter::px2mm(unsigned short x, unsigned short y, float& xmm, float& ymm) const {
    short detx = x - width_ / 2;
    short offset(detx / 256);
    switch (offset) {
        case -2: detx += -6;
            break;
        case -1: detx += -2;
            break;
        case 0: detx += 4;
            break;
        case 1: detx += 10;
            break;
        default:
            std::cout << "*** Error: Unknown Detector dimensions with detx = "
                    << detx << ",width = " << width_ << " and offset = " << offset << ".\n";
            throw std::runtime_error("Offset Calculation: unknown dorgx");
    }
    xmm = detx * Defs::Timepix::pixel_size_x;
    if (detx > 512) {
        xmm += Defs::Timepix::quad_gap_dx;
    }

    short dety = y - height_ / 2;
    offset = dety / 256;
    switch (offset) {
        case -2: dety += -6;
            break;
        case -1: dety += -2;
            break;
        case 0: dety += 4;
            break;
        case 1: dety += 10;
            break;
        default:
            std::cout << "*** Error: Unknown Detector dimensions.\n";
            throw std::runtime_error("Offset Calculation: unknown dorgy");
    }
    ymm = dety * Defs::Timepix::pixel_size_y;
    if (dety > 512) {
        ymm += Defs::Timepix::quad_gap_dy;
    }


}

/** Flag edge pixels as -1, which are each 175mu long: set to -1
 * w = 0; w = 255; w=256; w=511;
 * h = 0; h = 255; h=256; h=511
 */
void Converter::flag(Image<int>& image) {
        // set all (w'/h) to zero with w' as listed above
        for (size_t h = 0; h < image.height(); ++h) {
            size_t idx(0 + h * image.width());
            image.data(idx) = -1;
            idx = 255 + h * image.width();
            image.data(idx) = -1;
            idx = 256 + h * image.width();
            image.data(idx) = -1;
            idx = 511 + h * image.width();
            image.data(idx) = -1;
        }
        for (size_t w = 0; w < image.height(); ++w) {
            size_t idx(w + 0 * image.width());
            image.data(idx) = -1;
            idx = w + 255 * image.width();
            image.data(idx) = -1;
            idx = w + 256 * image.width();
            image.data(idx) = -1;
            idx = w + 511 * image.width();
            image.data(idx) = -1;
        }    
}

/**
 * Create input filename of the form inroot_i.bin
 * @param i
 * @return 
 */
std::string Converter::infile(unsigned short i) const {
    std::ostringstream outp;
    outp << inputdir_ << inroot_ << i << ".bin";
    return outp.str();
}

std::string Converter::outfile(const std::string& base, unsigned short i, unsigned short spacer) const{ 
    unsigned short digits (1);
    if (spacer == 0) {
        throw myExcepts::Usage ("Spacer must be > 0");
    }
    
    if (i > spacer) {
        digits = short(log10(i)+1);
    }
    else {
        digits = short(log10(spacer)+1);
    }
    std::ostringstream outp;
    outp << base;
    if (*(base.rbegin()) != '_') {
        outp << "_";
    }
    outp << std::setw(digits) << std::setfill('0') << i << ".cbf";
    return outp.str();
}

void Converter::print_missing(const std::vector<bool>& conv_flags) const {
    size_t idx (from_);
    
    while (idx != to_ && conv_flags[idx]) ++idx;
    if (idx == to_) return;
    
    std::cout << "Missing or ill-converted file numbers: \n";
    std::cout << idx << ' ';
    
    bool hyphen (false);
    for (size_t idx = from_; idx < to_; ++idx) {
        if (! conv_flags[idx]) {
            if (hyphen) continue;
            else {
                hyphen = true;
                std::cout << idx << '-';
                continue;
            }
        } else {
            if (!hyphen) continue;
            else {
                std::cout << idx - 1 << '\n';
                hyphen = false;
                continue;
            }
        }
    }
    if (conv_flags[to_])
        std::cout << to_;
    std::cout << std::endl;
}