/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

#include <iosfwd>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <algorithm>


#include <cbf/cbf.h>

#include "io.h"
#include "myExceptions.h"
#include "defines.h"

/**
 * Read in data from Timepix detector in binary format:
 * 15 bytes header, 512x512 or 1024x1024 bytes of data, unsigned short, little endian ('swab' for adxv)
 * @param filename
 * @param verbose print information about image, currently the maximum peak position
 * @param return the maximum peak position. 0, it verbose is false
 * @return 
 */
std::vector<unsigned short>  IO::in::raw(const char* filename, 
        unsigned short& width, unsigned short& height, size_t& peak, bool verbose) {
    std::ifstream inp(filename);
    if (!inp.is_open()) {
        throw myExcepts::FileMissing(filename);
    }

    // from Jan Pieter's 'FillSpace' Code. Header does not seem to be filled, so 
    // discard data;
    int arraysize;
    int headersize;
    int datadim;
    int datatype;

    char c1[5];

    // bytes 1-4
    inp.get(c1, 5);
    arraysize = c1[3] + (c1[2] << 8) + (c1[1] << 16) + (c1[0] << 24);
    // byte 5
    inp.get(c1, 2);
    headersize = c1[0];
    // byte 6
    inp.get(c1, 2);
    datadim = c1[0];
    // byte 7
    inp.get(c1, 2);
    datatype = c1[0];
    // bytes 8-11
    inp.get(c1, 5);
    height = c1[3] + (c1[2] << 8) + (int(c1[1]) << 16) + (c1[0] << 24);
    // bytes 12-15
    inp.get(c1, 5);
    width = c1[3] + (c1[2] << 8) + (int(c1[1]) << 16) + (c1[0] << 24);
    
    // read data - expect 512 x 512 unsigned shorts
    std::vector<unsigned short> data;
    data.reserve((int)height*(int)width);

    while (!inp.eof()) {
        int leftbit(inp.get());
        int rightbit(inp.get());
        unsigned short val = (leftbit << 8) + rightbit;
        if (inp.fail() || inp.eof()) {
            break;
        }
        data.push_back(val);
    }
    inp.close();

    if (verbose) {
        std::vector<unsigned short>::const_iterator it = std::max_element(data.begin(), data.end());
        peak = it - data.begin();
        int h = peak / height;
        int w = peak % height;
        std::cout << "Maximum value: " << *it << " at position (" << w << "/" 
                << h << ") for " << filename << "\n";
    }
    else {
        peak = 0;
    }

    return data;
}

std::vector<unsigned short>  IO::in::tiff(const char* filename, 
        unsigned short& width, unsigned short& height, bool verbose) {
    TIFF* tiff;
    int ret;
    uint16 bitsPerPixel;
    uint16 samplesPerPixel;

    TIFFSetWarningHandler(NULL);

    tiff = TIFFOpen(filename, "r");
    if (tiff == NULL) {
        printf("*** Error: Could not open \"%s\". Exiting.\n", filename);
        throw myExcepts::FileMissing(filename);
    }

    ret = TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &width);
    if (ret != 1) {
        printf("File Format Error: 'width' not declared in file header. Exiting\n");
        throw myExcepts::ImageFormat("Missing width in TIFF file");
    }
    ret = TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &height);
    if (ret != 1) {
        printf("File Format Error: 'height' not declared in file header. Exiting\n");
        throw myExcepts::ImageFormat("Missing height in TIFF file");
    }
    ret = TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerPixel);
    if (ret != 1) {
        printf("File Format Error: 'bitsPerPixel' not declared in file header. Exiting\n");
        throw myExcepts::ImageFormat("Missing pixel depth in TIFF file");
    }
    if (bitsPerPixel != 8 && bitsPerPixel != 16) {
        printf ("Error: Only TIFF images with 8 and 16 bits per pixel supported. Exiting\n");
        throw myExcepts::ImageFormat("Missing not (8 or 16) bits in TIFF file");
    }

    ret = TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &samplesPerPixel);
    if (ret != 1) {
        samplesPerPixel = 1;
    }

    printf("# Opening TIFF image \"%s\" %u x %u, %u bits per pixel\n",
            filename, width, height, bitsPerPixel);


    // reading data in strips; adapted from http://www.remotesensing.org/libtiff/libtiff.html#strips
    tsize_t scanlinelength = TIFFScanlineSize(tiff);
    tdata_t tiffstrip = _TIFFmalloc(scanlinelength);
        std::vector<unsigned short> data (width*height, 0);

    if (bitsPerPixel == 8) {
        for (int row = 0; row < height; ++row) {
            TIFFReadScanline(tiff, tiffstrip, row);
            for (int i = 0; i < width; ++i) {
                data[row * width + i] = ((char*) tiffstrip)[i];
            }
        }

    } else {
        for (int row = 0; row < height; ++row) {
            TIFFReadScanline(tiff, tiffstrip, row);
            for (int i = 0; i < width; ++i) {
                data[row * width + i] = ((uint16*) tiffstrip)[i];
            }
        }
    }

    _TIFFfree(tiffstrip);
    TIFFClose(tiff);

    return data;
}


/**
 * Write the data of the Image to a text file, in format 
 * x y I
 * @param filename
 * @param image
 * @return 
 */
void IO::out::txt(const char* filename, const Image<unsigned short>& image) {
    std::ofstream out(filename);

    if (!out.is_open()) {
        std::cout << "*** Error: Cannot open file " << filename << " for writing\n";
        throw myExcepts::FileMissing(filename);
    }

    std::vector <unsigned short> data(image.data());

    for (int h = 0; h < image.height(); ++h) {
        for (int w = 0; w < image.width(); ++w) {
            unsigned short val(data [w + h * image.width()]);
            out << std::setw(5) << w
                    << std::setw(5) << h
                    << std::setw(8) << val
                    << '\n';
        }
    }
    out.close();
}

/**
 * Write @c image to a simplified CBF file, readable by XDS
 * @param filename
 * @param image
 */
void IO::out::cbf(const char* filename, const Image<int>& image) {
    FILE* outp;
    outp = std::fopen(filename, "w");

   
    cbf_handle cbf;
    cbf_make_handle(&cbf);
    // create a CBF
    cbf_new_datablock(cbf, filename);

    // put a miniCBF header
    cbf_new_category(cbf, "array_data");
    cbf_new_column(cbf, "header_convention");
    cbf_set_value(cbf, "PILATUS_1.2");

    cbf_new_column(cbf, "header_contents");
    cbf_set_value(cbf, Defs::Timepix::cbf_header);

    // put the image
    cbf_new_category(cbf, "array_data");
    cbf_new_column(cbf, "data");
    int image_data [image.width() * image.height()];
    for (size_t i = 0; i < image.width() * image.height(); ++i) {
        image_data[i] = image.data(i);
    }
    cbf_set_integerarray_wdims_fs(cbf, CBF_BYTE_OFFSET, 1, image_data, sizeof (image_data[0]),
            1, image.width() * image.height(), "little_endian", image.width(),
            image.height(), 0, 1);

    cbf_write_widefile(cbf, outp, 1, CBF, MSG_DIGEST | MIME_HEADERS | PAD_4K, 0);
    //    fclose (outp);
    // no need to fclose() here as the 3rd argument "readable" is 1
    cbf_free_handle(cbf);
}


/**
 * Return a string containing the current date and time - 'ctime' reformatted
 * with the last carriage return removed
 */
std::string Utils::timestamp()
{
        std::time_t seconds = time(NULL);
        std::string now (std::ctime(&seconds));
        now.erase(now.end()-1);

        return now;
}
 

std::string Utils::cbfheader(const std::string detector, const std::string sn, 
        float pixel_x_mu, float pixel_y_mu,
        float silicon_thickness,
        float exposure,
        int cutoff,
        float threshold,
        int nexclpixels,
        const std::string& flatfield,
        float wavelength,
        float detdistance,
        float detvoffset,
        float beamx, float beamy,
        float alpha_start, float delta_alpha) {
    //24-0109-F
    std::ostringstream outp;
    outp << "# Detector: " << detector <<"S/N " << sn << "\n"
            << "# " << Utils::timestamp() << '\n'
            << "# Pixel_size "
            << std::scientific << std::setprecision(4) << pixel_x_mu << " m x "
            << std::scientific << std::setprecision(4) << pixel_y_mu << " m\n"
            <<  "# Silicon sensor, thickness " << silicon_thickness << "m\n"
            <<  "# Exposure_time " << exposure << "s\n"
            <<  "# Exposure_period 0.1000000 s\n"
            <<  "# Tau = 9.0e-2 s\n"
            <<  "# Count_cutoff " << cutoff << " counts\n"
            <<  "# Threshold_setting: " << threshold << " eV\n"
            <<  "# Gain_setting: low gain (vrf = -0.300)\n"
            <<  "# N_excluded_pixels = " << nexclpixels << "\n"
            <<  "# Excluded_pixels: badpix_mask.tif\n"
            <<  "# Flat_field: " << flatfield << "\n"
            <<  "# Trim_file: no_trim_file.bin\n"
            <<  "# Image_path: n/a\n"
            <<  "# Wavelength " << wavelength << " A\n"
            <<  "# Detector_distance " << detdistance << " m\n"
            <<  "# Detector_Voffset " << detvoffset << " m\n"
            <<  "# Beam_xy (" << beamx << ", " << beamy << ") pixels\n"
            <<  "# Alpha " << alpha_start << " deg\n"
            <<  "# Alpha_increment " << delta_alpha << " deg\n";
    return outp.str();
}